# -*- coding: utf-8 -*-
import pytest
import datetime
import copy

import numpy as np
import astropy.units as u
import matplotlib
try:
    from sunpy.visualization.animator import ImageAnimatorWCS, LineAnimator
except ImportError:
    from sunpy.visualization.imageanimator import ImageAnimatorWCS, LineAnimator

from ndcube import NDCube
from ndcube.utils.wcs import WCS
from ndcube.mixins import plotting


# sample data for tests
# TODO: use a fixture reading from a test file. file TBD.
ht = {'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
      'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
      'NAXIS2': 3,
      'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}
wt = WCS(header=ht, naxis=3)

hm = {'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
      'NAXIS1': 4,
      'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5,
      'NAXIS2': 3,
      'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm = WCS(header=hm, naxis=3)

data = np.array([[[1, 2, 3, 4], [2, 4, 5, 3], [0, -1, 2, 3]],
                 [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]]])
uncertainty = np.sqrt(data)
mask_cube = data < 0

cube = NDCube(
    data,
    wt,
    mask=mask_cube,
    uncertainty=uncertainty,
    missing_axis=[False, False, False, True],
    extra_coords=[('time', 0, u.Quantity(range(data.shape[0]), unit=u.s)),
                  ('hello', 1, u.Quantity(range(data.shape[1]), unit=u.W)),
                  ('bye', 2, u.Quantity(range(data.shape[2]), unit=u.m)),
                  ('another time', 2, np.array(
                      [datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                       for i in range(data.shape[2])])),
                  ('array coord', 2, np.arange(100, 100+data.shape[2]))
                  ])

cube_unit = NDCube(
    data,
    wt,
    mask=mask_cube,
    unit=u.J,
    uncertainty=uncertainty,
    missing_axis=[False, False, False, True],
    extra_coords=[('time', 0, u.Quantity(range(data.shape[0]), unit=u.s)),
                  ('hello', 1, u.Quantity(range(data.shape[1]), unit=u.W)),
                  ('bye', 2, u.Quantity(range(data.shape[2]), unit=u.m)),
                  ('another time', 2, np.array(
                      [datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                       for i in range(data.shape[2])]))
                  ])

cube_no_uncertainty = NDCube(
    data,
    wt,
    mask=mask_cube,
    missing_axis=[False, False, False, True],
    extra_coords=[('time', 0, u.Quantity(range(data.shape[0]), unit=u.s)),
                  ('hello', 1, u.Quantity(range(data.shape[1]), unit=u.W)),
                  ('bye', 2, u.Quantity(range(data.shape[2]), unit=u.m)),
                  ('another time', 2, np.array(
                      [datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                       for i in range(data.shape[2])]))
                  ])

cube_unit_no_uncertainty = NDCube(
    data,
    wt,
    mask=mask_cube,
    unit=u.J,
    missing_axis=[False, False, False, True],
    extra_coords=[('time', 0, u.Quantity(range(data.shape[0]), unit=u.s)),
                  ('hello', 1, u.Quantity(range(data.shape[1]), unit=u.W)),
                  ('bye', 2, u.Quantity(range(data.shape[2]), unit=u.m)),
                  ('another time', 2, np.array(
                      [datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                       for i in range(data.shape[2])]))
                  ])

cubem = NDCube(
    data,
    wm,
    mask=mask_cube,
    uncertainty=uncertainty,
    extra_coords=[('time', 0, u.Quantity(range(data.shape[0]), unit=u.s)),
                  ('hello', 1, u.Quantity(range(data.shape[1]), unit=u.W)),
                  ('bye', 2, u.Quantity(range(data.shape[2]), unit=u.m)),
                  ('another time', 2, np.array(
                      [datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                       for i in range(data.shape[2])]))
                  ])

# Derive expected data values
cube_data = np.ma.masked_array(cube.data, cube.mask)

# Derive expected axis_ranges.
# Let False stand for ranges generated by SunPy classes and so not tested.
cube_none_axis_ranges_axis2 = [False, False, np.array([0.4, 0.8, 1.2, 1.6])]

cube_none_axis_ranges_axis2_s = copy.deepcopy(cube_none_axis_ranges_axis2)
cube_none_axis_ranges_axis2_s[2] = cube_none_axis_ranges_axis2[2] * 60.

cube_none_axis_ranges_axis2_bye = copy.deepcopy(cube_none_axis_ranges_axis2)
cube_none_axis_ranges_axis2_bye[2] = cube.extra_coords["bye"]["value"].value

cube_none_axis_ranges_axis2_array = copy.deepcopy(cube_none_axis_ranges_axis2)
cube_none_axis_ranges_axis2_array[2] = np.arange(10, 10+cube.data.shape[-1])

@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (cube[0, 0], {},
     (np.ma.masked_array([0.4, 0.8, 1.2, 1.6], cube[0, 0].mask),
      np.ma.masked_array(cube[0, 0].data, cube[0, 0].mask),
      "time [min]", "Data [None]", (0.4, 1.6), (1, 4))),

    (cube_unit[0, 0], {"axes_coordinates": "bye", "axes_units": "km", "data_unit": u.erg},
     (np.ma.masked_array(cube_unit[0, 0].extra_coords["bye"]["value"].to(u.km).value,
                         cube_unit[0, 0].mask),
      np.ma.masked_array(u.Quantity(cube_unit[0, 0].data,
                                    unit=cube_unit[0, 0].unit).to(u.erg).value,
                         cube_unit[0, 0].mask),
      "bye [km]", "Data [erg]", (0, 0.003), (10000000, 40000000))),

    (cube_unit[0, 0], {"axes_coordinates": np.arange(10, 10+cube_unit[0, 0].data.shape[0])},
     (np.ma.masked_array(np.arange(10, 10+cube_unit[0, 0].data.shape[0]), cube_unit[0, 0].mask),
      np.ma.masked_array(cube_unit[0, 0].data, cube_unit[0, 0].mask),
      " [None]", "Data [J]", (10, 10+cube_unit[0, 0].data.shape[0]-1), (1, 4))),

    (cube_no_uncertainty[0, 0], {},
     (np.ma.masked_array([0.4, 0.8, 1.2, 1.6], cube_no_uncertainty[0, 0].mask),
      np.ma.masked_array(cube_no_uncertainty[0, 0].data, cube_no_uncertainty[0, 0].mask),
      "time [min]", "Data [None]", (0.4, 1.6), (1, 4))),

    (cube_unit_no_uncertainty[0, 0], {},
     (np.ma.masked_array([0.4, 0.8, 1.2, 1.6], cube_unit_no_uncertainty[0, 0].mask),
      np.ma.masked_array(cube_no_uncertainty[0, 0].data, cube_unit_no_uncertainty[0, 0].mask),
      "time [min]", "Data [J]", (0.4, 1.6), (1, 4))),

    (cube_unit_no_uncertainty[0, 0], {"data_unit": u.erg},
     (np.ma.masked_array([0.4, 0.8, 1.2, 1.6], cube_unit_no_uncertainty[0, 0].mask),
      np.ma.masked_array(u.Quantity(cube_unit[0, 0].data,
                                    unit=cube_unit[0, 0].unit).to(u.erg).value,
                         cube_unit[0, 0].mask),
      "time [min]", "Data [erg]", (0.4, 1.6), (10000000, 40000000)))
    ])
def test_cube_plot_1D(test_input, test_kwargs, expected_values):
    # Unpack expected properties.
    expected_xdata, expected_ydata, expected_xlabel, expected_ylabel, \
      expected_xlim, expected_ylim = expected_values
    # Run plot method.
    output = test_input.plot(**test_kwargs)
    # Check plot properties are correct.
    # Type
    assert isinstance(output, matplotlib.axes.Axes)
    # Check x axis data
    output_xdata = (output.axes.lines[0].get_xdata())
    assert np.allclose(output_xdata.data, expected_xdata.data)
    if isinstance(output_xdata.mask, np.ndarray):
        np.testing.assert_array_equal(output_xdata.mask, expected_xdata.mask)
    else:
        assert output_xdata.mask == expected_xdata.mask
    # Check y axis data
    output_ydata = (output.axes.lines[0].get_ydata())
    assert np.allclose(output_ydata.data, expected_ydata.data)
    if isinstance(output_ydata.mask, np.ndarray):
        np.testing.assert_array_equal(output_ydata.mask, expected_ydata.mask)
    else:
        assert output_ydata.mask == expected_ydata.mask
    # Check axis labels
    assert output.axes.get_xlabel() == expected_xlabel
    assert output.axes.get_ylabel() == expected_ylabel
    # Check axis limits
    output_xlim = output.axes.get_xlim()
    assert output_xlim[0] <= expected_xlim[0]
    assert output_xlim[1] >= expected_xlim[1]
    output_ylim = output.axes.get_ylim()
    assert output_ylim[0] <= expected_ylim[0]
    assert output_ylim[1] >= expected_ylim[1]


@pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
    (cube[0, 0], {"axes_coordinates": np.arange(10, 10+cube_unit[0, 0].data.shape[0]),
                  "axes_units": u.C}, TypeError),
    (cube[0, 0], {"data_unit": u.C}, TypeError)
    ])
def test_cube_plot_1D_errors(test_input, test_kwargs, expected_error):
    with pytest.raises(expected_error):
        output = test_input.plot(**test_kwargs)


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (cube[0], {},
     (np.ma.masked_array(cube[0].data, cube[0].mask), "time [min]", "em.wl [m]",
      (0.4, 1.6, 2e-11, 6e-11))),

    (cube[0], {"axes_coordinates": ["bye", None], "axes_units": [None, u.cm]},
     (np.ma.masked_array(cube[0].data, cube[0].mask), "bye [m]", "em.wl [cm]",
      (0.0, 3.0, 2e-9, 6e-9))),

    (cube[0], {"axes_coordinates": [np.arange(10, 10+cube[0].data.shape[1]),
                                    u.Quantity(np.arange(10, 10+cube[0].data.shape[0]), unit=u.m)],
               "axes_units": [None, u.cm]},
     (np.ma.masked_array(cube[0].data, cube[0].mask), " [None]", " [cm]", (10, 13, 1000, 1200))),

    (cube[0], {"axes_coordinates": [np.arange(10, 10+cube[0].data.shape[1]),
                                    u.Quantity(np.arange(10, 10+cube[0].data.shape[0]), unit=u.m)]},
     (np.ma.masked_array(cube[0].data, cube[0].mask), " [None]", " [m]", (10, 13, 10, 12))),

    (cube_unit[0], {"plot_axis_indices": [0, 1], "axes_coordinates": [None, "bye"],
                    "data_unit": u.erg},
     (np.ma.masked_array((cube_unit[0].data * cube_unit[0].unit).to(u.erg).value,
                         cube_unit[0].mask).transpose(),
      "em.wl [m]", "bye [m]", (2e-11, 6e-11, 0.0, 3.0)))
    ])
def test_cube_plot_2D(test_input, test_kwargs, expected_values):
    # Unpack expected properties.
    expected_data, expected_xlabel, expected_ylabel, expected_extent = \
      expected_values
    # Run plot method.
    output = test_input.plot(**test_kwargs)
    # Check plot properties are correct.
    assert isinstance(output, matplotlib.axes.Axes)
    np.testing.assert_array_equal(output.images[0].get_array(), expected_data)
    assert output.axes.xaxis.get_label_text() == expected_xlabel
    assert output.axes.yaxis.get_label_text() == expected_ylabel
    assert np.allclose(output.images[0].get_extent(), expected_extent)


@pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
    (cube[0], {"axes_coordinates": ["array coord", None], "axes_units": [u.cm, None]}, TypeError),
    (cube[0], {"axes_coordinates": [np.arange(10, 10+cube[0].data.shape[1]), None],
               "axes_units": [u.cm, None]}, TypeError),
    (cube[0], {"data_unit": u.cm}, TypeError)
    ])
def test_cube_plot_2D_errors(test_input, test_kwargs, expected_error):
    with pytest.raises(expected_error):
        output = test_input.plot(**test_kwargs)


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (cubem, {},
     (cubem.data, [np.array([0., 2.]), [0, 3], [0, 4]], "", ""))
    ])
def test_cube_plot_ND_as_2DAnimation(test_input, test_kwargs, expected_values):
    # Unpack expected properties.
    expected_data, expected_axis_ranges, expected_xlabel, expected_ylabel = expected_values
    # Run plot method.
    output = test_input.plot(**test_kwargs)
    # Check plot properties are correct.
    assert type(output) is ImageAnimatorWCS
    np.testing.assert_array_equal(output.data, expected_data)
    assert output.axes.xaxis.get_label_text() == expected_xlabel
    assert output.axes.yaxis.get_label_text() == expected_ylabel


@pytest.mark.parametrize("input_values, expected_values", [
    ((None, None, None, None, {"image_axes": [-1, -2],
                               "axis_ranges": [np.arange(3), np.arange(3)],
                               "unit_x_axis": "km",
                               "unit_y_axis": u.s,
                               "unit": u.W}),
     ([-1, -2], [np.arange(3), np.arange(3)], ["km", u.s], u.W, {})),
    (([-1, -2], [np.arange(3), np.arange(3)], ["km", u.s], u.W, {}),
     ([-1, -2], [np.arange(3), np.arange(3)], ["km", u.s], u.W, {})),
    (([-1], None, None, None, {"unit_x_axis": "km"}),
     ([-1], None, "km", None, {})),
    (([-1, -2], None, None, None, {"unit_x_axis": "km"}),
     (([-1, -2], None, ["km", None], None, {}))),
    (([-1, -2], None, None, None, {"unit_y_axis": "km"}),
     (([-1, -2], None, [None, "km"], None, {})))
    ])
def test_support_101_plot_API(input_values, expected_values):
    # Define expected values.
    expected_plot_axis_indices, expected_axes_coordinates, expected_axes_units, \
      expected_data_unit, expected_kwargs = expected_values
    # Run function
    output_plot_axis_indices, output_axes_coordinates, output_axes_units, \
      output_data_unit, output_kwargs = plotting._support_101_plot_API(*input_values)
    # Check values are correct
    assert output_plot_axis_indices == expected_plot_axis_indices
    if expected_axes_coordinates is None:
        assert output_axes_coordinates == expected_axes_coordinates
    elif type(expected_axes_coordinates) is list:
        for i, ac in enumerate(output_axes_coordinates):
            np.testing.assert_array_equal(ac, expected_axes_coordinates[i])
    assert output_axes_units == expected_axes_units
    assert output_data_unit == expected_data_unit
    assert output_kwargs == expected_kwargs


@pytest.mark.parametrize("input_values", [
    ([0, 1], None, None, None, {"image_axes": [-1, -2]}),
    (None, [np.arange(1, 4), np.arange(1, 4)], None, None,
      {"axis_ranges": [np.arange(3), np.arange(3)]}),
    (None, None, [u.s, "km"], None, {"unit_x_axis": u.W}),
    (None, None, [u.s, "km"], None, {"unit_y_axis": u.W}),
    (None, None, None, u.s, {"unit": u.W}),
    ([0, 1, 2], None, None, None, {"unit_x_axis": [u.s, u.km, u.W]}),
    ])
def test_support_101_plot_API_errors(input_values):
    with pytest.raises(ValueError):
        output = plotting._support_101_plot_API(*input_values)


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (cube, {"plot_axis_indices": -1},
     (cube_data, cube_none_axis_ranges_axis2, "time [min]", "Data [None]")),

    (cube_unit, {"plot_axis_indices": -1, "axes_units": u.s, "data_unit": u.erg},
     (cube_data*1e7, cube_none_axis_ranges_axis2_s, "time [s]", "Data [erg]")),

    (cube_unit, {"plot_axis_indices": -1, "axes_coordinates": "bye"},
     (cube_data, cube_none_axis_ranges_axis2_bye, "bye [m]", "Data [J]")),

    (cube, {"plot_axis_indices": -1, "axes_coordinates": np.arange(10, 10+cube.data.shape[-1])},
     (cube_data, cube_none_axis_ranges_axis2_array, " [None]", "Data [None]"))
    ])
def test_cube_plot_ND_as_1DAnimation(test_input, test_kwargs, expected_values):
    # Unpack expected properties.
    expected_data, expected_axis_ranges, expected_xlabel, expected_ylabel = expected_values
    # Run plot method.
    output = test_input.plot(**test_kwargs)
    # Check plot properties are correct.
    assert type(output) is LineAnimator
    np.testing.assert_array_equal(output.data, expected_data)
    for i, output_axis_range in enumerate(output.axis_ranges):
        if expected_axis_ranges[i] is not False:
            assert np.allclose(output_axis_range, expected_axis_ranges[i])
    assert output.axes.xaxis.get_label_text() == expected_xlabel
    assert output.axes.yaxis.get_label_text() == expected_ylabel



# -*- coding: utf-8 -*-

"""
Helpers for testing ndcube.
"""
import unittest

import numpy as np

from ndcube import utils

__all__ = ['assert_extra_coords_equal',
           'assert_metas_equal',
           'assert_cubes_equal',
           'assert_cubesequences_equal',
           'assert_wcs_are_equal']


def assert_extra_coords_equal(test_input, extra_coords):
    assert test_input.keys() == extra_coords.keys()
    for key in list(test_input.keys()):
        assert test_input[key]['axis'] == extra_coords[key]['axis']
        assert (test_input[key]['value'] == extra_coords[key]['value']).all()


def assert_metas_equal(test_input, expected_output):
    assert test_input.keys() == expected_output.keys()
    for key in list(test_input.keys()):
        assert test_input[key] == expected_output[key]


def assert_cubes_equal(test_input, expected_cube):
    unit_tester = unittest.TestCase()
    assert type(test_input) == type(expected_cube)
    assert np.all(test_input.mask == expected_cube.mask)
    assert_wcs_are_equal(test_input.wcs, expected_cube.wcs)
    assert test_input.missing_axis == expected_cube.missing_axis
    assert test_input.uncertainty.array.shape == expected_cube.uncertainty.array.shape
    assert test_input.world_axis_physical_types == expected_cube.world_axis_physical_types
    assert all(test_input.dimensions.value == expected_cube.dimensions.value)
    assert test_input.dimensions.unit == expected_cube.dimensions.unit
    assert_extra_coords_equal(test_input.extra_coords, expected_cube.extra_coords)


def assert_cubesequences_equal(test_input, expected_sequence):
    assert type(test_input) == type(expected_sequence)
    assert_metas_equal(test_input.meta, expected_sequence.meta)
    assert test_input._common_axis == expected_sequence._common_axis
    for i, cube in enumerate(test_input.data):
        assert_cubes_equal(cube, expected_sequence.data[i])


def assert_wcs_are_equal(wcs1, wcs2):
    """
    Assert function for testing two wcs object.
    Used in testing NDCube.
    """
    assert list(wcs1.wcs.ctype) == list(wcs2.wcs.ctype)
    assert list(wcs1.wcs.crval) == list(wcs2.wcs.crval)
    assert list(wcs1.wcs.crpix) == list(wcs2.wcs.crpix)
    assert list(wcs1.wcs.cdelt) == list(wcs2.wcs.cdelt)
    assert list(wcs1.wcs.cunit) == list(wcs2.wcs.cunit)
    assert wcs1.wcs.naxis == wcs2.wcs.naxis



# -*- coding: utf-8 -*-
import pytest
import unittest

import numpy as np
import astropy.units as u

from ndcube import utils

missing_axis_none = [False]*3
missing_axis_0_2 = [True, False, True]
missing_axis_1 = [False, True, False]

axes_length = 3
extra_coords_dict = {"time": {"axis": 0, "value": u.Quantity(range(axes_length), unit=u.pix)},
                     "hello": {"axis": 1, "value": u.Quantity(range(axes_length), unit=u.pix)}}
extra_coords_input = [('time', 0, u.Quantity(range(axes_length), unit=u.pix)),
                      ('hello', 1, u.Quantity(range(axes_length), unit=u.pix))]
extra_coords_dict_wcs = {"time": {"wcs axis": 0,
                                  "value": u.Quantity(range(axes_length), unit=u.pix)},
                         "hello": {"wcs axis": 1,
                                   "value": u.Quantity(range(axes_length), unit=u.pix)}}


@pytest.mark.parametrize(
    "test_input,expected",
    [((None, missing_axis_none), None),
     ((0, missing_axis_none), 2),
     ((1, missing_axis_none), 1),
     ((0, missing_axis_0_2), 1),
     ((1, missing_axis_1), 0),
     ((-1, missing_axis_0_2), 1),
     ((-2, missing_axis_1), 2),
     ((-1, missing_axis_none), 0)])
def test_data_axis_to_wcs_axis(test_input, expected):
    assert utils.cube.data_axis_to_wcs_axis(*test_input) == expected


@pytest.mark.parametrize("test_input", [(-2, missing_axis_0_2), (1, missing_axis_0_2)])
def test_data_axis_to_wcs_axis_error(test_input):
    with pytest.raises(IndexError):
        utils.cube.data_axis_to_wcs_axis(*test_input)


@pytest.mark.parametrize(
    "test_input,expected",
    [((None, missing_axis_none), None),
     ((0, missing_axis_none), 2),
     ((1, missing_axis_none), 1),
     ((1, missing_axis_0_2), 0),
     ((0, missing_axis_1), 1),
     ((-1, missing_axis_0_2), None),
     ((-2, missing_axis_0_2), 0),
     ((-2, missing_axis_1), None),
     ((-3, missing_axis_1), 1),
     ((-1, missing_axis_none), 0)])
def test_wcs_axis_to_data_axis(test_input, expected):
    assert utils.cube.wcs_axis_to_data_axis(*test_input) == expected


@pytest.mark.parametrize("test_input", [(-10, missing_axis_0_2), (10, missing_axis_0_2)])
def test_wcs_axis_to_data_axis_error(test_input):
    with pytest.raises(IndexError):
        utils.cube.data_axis_to_wcs_axis(*test_input)


def test_select_order():
    lists = [['TIME', 'WAVE', 'HPLT-TAN',
              'HPLN-TAN'], ['WAVE', 'HPLT-TAN', 'UTC',
                            'HPLN-TAN'], ['HPLT-TAN', 'TIME', 'HPLN-TAN'],
             ['HPLT-TAN', 'DEC--TAN',
              'WAVE'], [], ['UTC', 'TIME', 'WAVE', 'HPLT-TAN']]

    results = [
        [0, 1, 2, 3],
        [2, 0, 1, 3],
        [1, 0, 2],  # Second order is initial order
        [2, 0, 1],
        [],
        [1, 0, 2, 3]
    ]

    for (l, r) in zip(lists, results):
        assert utils.cube.select_order(l) == r


@pytest.mark.parametrize("test_input", [
    ([('name', 0)], [False, False], (1, 2)),
    ([(0, 0, 0)], [False, False], (1, 2)),
    ([('name', '0', 0)], [False, False], (1, 2)),
    ([('name', 0, [0, 1])], [False, False], (1, 2))
    ])
def test_format_input_extra_coords_to_extra_coords_wcs_axis_value(test_input):
    with pytest.raises(ValueError):
        utils.cube._format_input_extra_coords_to_extra_coords_wcs_axis(*test_input)


@pytest.mark.parametrize("test_input,expected", [
    ((extra_coords_dict, missing_axis_none), extra_coords_input),

    ((extra_coords_dict_wcs,  missing_axis_none),
     [('time', 2, u.Quantity(range(axes_length), unit=u.pix)),
      ('hello', 1, u.Quantity(range(axes_length), unit=u.pix))]),

    ((extra_coords_dict_wcs,  missing_axis_1),
     [('time', 1, u.Quantity(range(axes_length), unit=u.pix)),
      ('hello', None, u.Quantity(range(axes_length), unit=u.pix))])
    ])
def test_convert_extra_coords_dict_to_input_format(test_input, expected):
    output = utils.cube.convert_extra_coords_dict_to_input_format(*test_input)
    if len(output) != len(expected):
        raise AssertionError("{0} != {1}".format(output, expected))
    for output_tuple in output:
        j = 0
        while j < len(expected):
            if output_tuple[0] == expected[j][0]:
                assert len(output_tuple) == len(expected[j])
                print(output_tuple)
                print(expected[j])
                for k, el in enumerate(output_tuple):
                    try:
                        assert el == expected[j][k]
                    except ValueError as err:
                        if err.args[0] == "The truth value of an array with more than" + \
                          " one element is ambiguous. Use a.any() or a.all()":
                            assert (el == expected[j][k]).all()
                        else:
                            raise err
                j = len(expected)+1
            else:
                j += 1
        if j == len(expected):
            raise AssertionError("{0} != {1}".format(output, expected))


def test_convert_extra_coords_dict_to_input_format_error():
    with pytest.raises(KeyError):
        utils.cube.convert_extra_coords_dict_to_input_format(
            {"time": {"not axis": 0, "value": []}}, missing_axis_none)



import copy

from astropy.nddata.mixins.ndslicing import NDSlicingMixin

from ndcube import utils

__all__ = ['NDCubeSlicingMixin']


class NDCubeSlicingMixin(NDSlicingMixin):
    # Inherit docstring from parent class
    __doc__ = NDSlicingMixin.__doc__

    def _slice_wcs(self, item):
        """
        Override parent class method so we disable the wcs slicing on
        `astropy.nddata.mixins.NDSlicingMixin`.
        """
        return None

    def __getitem__(self, item):
        """
        Override the parent class method to explicitly catch `None` indices.

        This method calls ``_slice`` and then constructs a new object using the
        kwargs returned by ``_slice``.
        """
        if item is None or (isinstance(item, tuple) and None in item):
            raise IndexError("None indices not supported")

        return super().__getitem__(item)

    def _slice(self, item):
        """
        Construct a set of keyword arguments to initialise a new (sliced)
        instance of the class. This method is called in
        `astropy.nddata.mixins.NDSlicingMixin.__getitem__`.

        This method extends the `~astropy.nddata.mixins.NDSlicingMixin` method
        to add support for ``missing_axis`` and ``extra_coords`` and overwrites
        the astropy handling of wcs slicing.
        """
        kwargs = super()._slice(item)

        wcs, missing_axis = self._slice_wcs_missing_axis(item)
        kwargs['wcs'] = wcs
        kwargs['missing_axis'] = missing_axis
        kwargs['extra_coords'] = self._slice_extra_coords(item, missing_axis)

        return kwargs

    def _slice_wcs_missing_axis(self, item):
        # here missing axis is reversed as the item comes already in the reverse order
        # of the input
        return utils.wcs._wcs_slicer(
            self.wcs, copy.deepcopy(self.missing_axis[::-1]), item)

    def _slice_extra_coords(self, item, missing_axis):
        if self.extra_coords is None:
            new_extra_coords_dict = None
        else:
            old_extra_coords = self.extra_coords
            extra_coords_keys = list(old_extra_coords.keys())
            new_extra_coords = copy.deepcopy(self._extra_coords_wcs_axis)
            for ck in extra_coords_keys:
                axis_ck = old_extra_coords[ck]["axis"]
                if isinstance(item, (slice, int)):
                    if axis_ck == 0:
                        new_extra_coords[ck]["value"] = new_extra_coords[ck]["value"][item]
                if isinstance(item, tuple):
                    try:
                        slice_item_extra_coords = item[axis_ck]
                        new_extra_coords[ck]["value"] = \
                            new_extra_coords[ck]["value"][slice_item_extra_coords]
                    except IndexError:
                        pass
                    except TypeError:
                        pass
                new_extra_coords_dict = utils.cube.convert_extra_coords_dict_to_input_format(
                    new_extra_coords, missing_axis)
        return new_extra_coords_dict



import datetime
from warnings import warn

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import astropy.units as u
import sunpy.visualization.wcsaxes_compat as wcsaxes_compat
try:
    from sunpy.visualization.animator import ImageAnimator, ImageAnimatorWCS, LineAnimator
except ImportError:
    from sunpy.visualization.imageanimator import ImageAnimator, ImageAnimatorWCS, LineAnimator

from ndcube import utils
from ndcube.mixins import sequence_plotting

__all__ = ['NDCubePlotMixin']

INVALID_UNIT_SET_MESSAGE = "Can only set unit for axis if corresponding coordinates in " + \
  "axes_coordinates are set to None, an astropy Quantity or the name of an extra coord that " + \
  "is an astropy Quantity."


class NDCubePlotMixin:
    """
    Add plotting functionality to a NDCube class.
    """

    def plot(self, axes=None, plot_axis_indices=None, axes_coordinates=None,
             axes_units=None, data_unit=None, **kwargs):
        """
        Plots an interactive visualization of this cube with a slider
        controlling the wavelength axis for data having dimensions greater than 2.
        Plots an x-y graph onto the current axes for 2D or 1D data.
        Keyword arguments are passed on to matplotlib.
        Parameters other than data and wcs are passed to ImageAnimatorWCS,
        which in turn passes them to imshow for data greater than 2D.

        Parameters
        ----------
        plot_axis_indices: `list`
            The two axes that make the image.
            Default=[-1,-2].  This implies cube instance -1 dimension
            will be x-axis and -2 dimension will be y-axis.

        axes: `astropy.visualization.wcsaxes.core.WCSAxes` or None:
            The axes to plot onto. If None the current axes will be used.

        axes_unit: `list` of `astropy.units.Unit`

        data_unit: `astropy.unit.Unit`
            The data is changed to the unit given or the cube.unit if not given, for 1D plots.

        axes_coordinates: list of physical coordinates for array or None
            If None array indices will be used for all axes.
            If a list it should contain one element for each axis of the numpy array.
            For the image axes a [min, max] pair should be specified which will be
            passed to :func:`matplotlib.pyplot.imshow` as extent.
            For the slider axes a [min, max] pair can be specified or an array the
            same length as the axis which will provide all values for that slider.
            If None is specified for an axis then the array indices will be used
            for that axis.

        """
        # If old API is used, convert to new API.
        plot_axis_indices, axes_coordiantes, axes_units, data_unit, kwargs = _support_101_plot_API(
            plot_axis_indices, axes_coordinates, axes_units, data_unit, kwargs)
        # Check kwargs are in consistent formats and set default values if not done so by user.
        naxis = len(self.dimensions)
        plot_axis_indices, axes_coordinates, axes_units = sequence_plotting._prep_axes_kwargs(
            naxis, plot_axis_indices, axes_coordinates, axes_units)
        if naxis is 1:
            ax = self._plot_1D_cube(axes, axes_coordinates,
                                    axes_units, data_unit, **kwargs)
        else:
            if len(plot_axis_indices) == 1:
                ax = self._animate_cube_1D(
                    plot_axis_index=plot_axis_indices[0], axes_coordinates=axes_coordinates,
                    axes_units=axes_units, data_unit=data_unit, **kwargs)
            else:
                if naxis == 2:
                    ax = self._plot_2D_cube(axes, plot_axis_indices, axes_coordinates,
                                            axes_units, data_unit, **kwargs)
                else:
                    ax = self._plot_3D_cube(
                        plot_axis_indices=plot_axis_indices, axes_coordinates=axes_coordinates,
                        axes_units=axes_units, **kwargs)
        return ax

    def _plot_1D_cube(self, axes=None, axes_coordinates=None, axes_units=None, data_unit=None,
                      **kwargs):
        """
        Plots a graph.
        Keyword arguments are passed on to matplotlib.

        Parameters
        ----------
        data_unit: `astropy.unit.Unit`
            The data is changed to the unit given or the cube.unit if not given.

        """
        # Derive x-axis coordinates and unit from inputs.
        x_axis_coordinates, unit_x_axis = sequence_plotting._derive_1D_coordinates_and_units(
            axes_coordinates, axes_units)
        if x_axis_coordinates is None:
            # Default is to derive x coords and defaul xlabel from WCS object.
            xname = self.world_axis_physical_types[0]
            xdata = self.axis_world_coords()
        elif isinstance(x_axis_coordinates, str):
            # User has entered a str as x coords, get that extra coord.
            xname = x_axis_coordinates
            xdata = self.extra_coords[x_axis_coordinates]["value"]
        else:
            # Else user must have set the x-values manually.
            xname = ""
            xdata = x_axis_coordinates
        # If a unit has been set for the x-axis, try to convert x coords to that unit.
        if isinstance(xdata, u.Quantity):
            if unit_x_axis is None:
                unit_x_axis = xdata.unit
                xdata = xdata.value
            else:
                xdata = xdata.to(unit_x_axis).value
        else:
            if unit_x_axis is not None:
                raise TypeError(INVALID_UNIT_SET_MESSAGE)
        # Define default x axis label.
        default_xlabel = "{0} [{1}]".format(xname, unit_x_axis)
        # Combine data and uncertainty with mask.
        xdata = np.ma.masked_array(xdata, self.mask)
        # Derive y-axis coordinates, uncertainty and unit from the NDCube's data.
        if self.unit is None:
            if data_unit is not None:
                raise TypeError("Can only set y-axis unit if self.unit is set to a "
                                "compatible unit.")
            else:
                ydata = self.data
                if self.uncertainty is None:
                    yerror = None
                else:
                    yerror = self.uncertainty.array
        else:
            if data_unit is None:
                data_unit = self.unit
                ydata = self.data
                if self.uncertainty is None:
                    yerror = None
                else:
                    yerror = self.uncertainty.array
            else:
                ydata = (self.data * self.unit).to(data_unit).value
                if self.uncertainty is None:
                    yerror = None
                else:
                    yerror = (self.uncertainty.array * self.unit).to(data_unit).value
        # Combine data and uncertainty with mask.
        ydata = np.ma.masked_array(ydata, self.mask)
        if yerror is not None:
            yerror = np.ma.masked_array(yerror, self.mask)
        # Create plot
        fig, ax = sequence_plotting._make_1D_sequence_plot(xdata, ydata, yerror,
                                                           data_unit, default_xlabel, kwargs)
        return ax

    def _plot_2D_cube(self, axes=None, plot_axis_indices=None, axes_coordinates=None,
                      axes_units=None, data_unit=None, **kwargs):
        """
        Plots a 2D image onto the current
        axes. Keyword arguments are passed on to matplotlib.

        Parameters
        ----------
        axes: `astropy.visualization.wcsaxes.core.WCSAxes` or `None`:
            The axes to plot onto. If None the current axes will be used.

        plot_axis_indices: `list`.
            The first axis in WCS object will become the first axis of plot_axis_indices and
            second axis in WCS object will become the second axis of plot_axis_indices.
            Default: ['x', 'y']

        """
        # Set default values of kwargs if not set.
        if axes_coordinates is None:
            axes_coordinates = [None, None]
        if axes_units is None:
            axes_units = [None, None]
        # Set which cube dimensions are on the x an y axes.
        axis_data = ['x', 'x']
        axis_data[plot_axis_indices[1]] = 'y'
        axis_data = axis_data[::-1]
        # Determine data to be plotted
        if data_unit is None:
            data = self.data
        else:
            # If user set data_unit, convert dat to desired unit if self.unit set.
            if self.unit is None:
                raise TypeError("Can only set data_unit if NDCube.unit is set.")
            else:
                data = (self.data * self.unit).to(data_unit).value
        # Combine data with mask
        data = np.ma.masked_array(data, self.mask)
        if axes is None:
            try:
                axes_coord_check = axes_coordinates == [None, None]
            except:
                axes_coord_check = False
            if axes_coord_check:
                # Build slice list for WCS for initializing WCSAxes object.
                if self.wcs.naxis != 2:
                    slice_list = []
                    index = 0
                    for bool_ in self.missing_axis:
                        if not bool_:
                            slice_list.append(axis_data[index])
                            index += 1
                        else:
                            slice_list.append(1)
                    if index != 2:
                        raise ValueError("Dimensions of WCS and data don't match")
                    ax = wcsaxes_compat.gca_wcs(self.wcs, slices=tuple(slice_list))
                else:
                    ax = wcsaxes_compat.gca_wcs(self.wcs)
                # Set axis labels
                x_wcs_axis = utils.cube.data_axis_to_wcs_axis(plot_axis_indices[0],
                                                              self.missing_axis)
                ax.set_xlabel("{0} [{1}]".format(
                    self.world_axis_physical_types[plot_axis_indices[0]],
                    self.wcs.wcs.cunit[x_wcs_axis]))
                y_wcs_axis = utils.cube.data_axis_to_wcs_axis(plot_axis_indices[1],
                                                              self.missing_axis)
                ax.set_ylabel("{0} [{1}]".format(
                    self.world_axis_physical_types[plot_axis_indices[1]],
                    self.wcs.wcs.cunit[y_wcs_axis]))
                # Plot data
                ax.imshow(data, **kwargs)
            else:
                # Else manually set axes x and y values based on user's input for axes_coordinates.
                new_axes_coordinates, new_axis_units, default_labels = \
                  self._derive_axes_coordinates(axes_coordinates, axes_units)
                # Initialize axes object and set values along axis.
                fig, ax = plt.subplots(1, 1)
                # Since we can't assume the x-axis will be uniform, create NonUniformImage
                # axes and add it to the axes object.
                if plot_axis_indices[0] < plot_axis_indices[1]:
                    data = data.transpose()
                im_ax = mpl.image.NonUniformImage(
                    ax, extent=(new_axes_coordinates[plot_axis_indices[0]][0],
                                new_axes_coordinates[plot_axis_indices[0]][-1],
                                new_axes_coordinates[plot_axis_indices[1]][0],
                                new_axes_coordinates[plot_axis_indices[1]][-1]), **kwargs)
                im_ax.set_data(new_axes_coordinates[plot_axis_indices[0]],
                               new_axes_coordinates[plot_axis_indices[1]], data)
                ax.add_image(im_ax)
                # Set the limits, labels, etc. of the axes.
                xlim = kwargs.pop("xlim", (new_axes_coordinates[plot_axis_indices[0]][0],
                                           new_axes_coordinates[plot_axis_indices[0]][-1]))
                ax.set_xlim(xlim)
                ylim = kwargs.pop("xlim", (new_axes_coordinates[plot_axis_indices[1]][0],
                                           new_axes_coordinates[plot_axis_indices[1]][-1]))
                ax.set_ylim(ylim)
                xlabel = kwargs.pop("xlabel", default_labels[plot_axis_indices[0]])
                ylabel = kwargs.pop("ylabel", default_labels[plot_axis_indices[1]])
                ax.set_xlabel(xlabel)
                ax.set_ylabel(ylabel)
        return ax

    def _plot_3D_cube(self, plot_axis_indices=None, axes_coordinates=None,
                      axes_units=None, data_unit=None, **kwargs):
        """
        Plots an interactive visualization of this cube using sliders to move through axes
        plot using in the image.
        Parameters other than data and wcs are passed to ImageAnimatorWCS, which in turn
        passes them to imshow.

        Parameters
        ----------
        plot_axis_indices: `list`
            The two axes that make the image.
            Like [-1,-2] this implies cube instance -1 dimension
            will be x-axis and -2 dimension will be y-axis.

        axes_unit: `list` of `astropy.units.Unit`

        axes_coordinates: `list` of physical coordinates for array or None
            If None array indices will be used for all axes.
            If a list it should contain one element for each axis of the numpy array.
            For the image axes a [min, max] pair should be specified which will be
            passed to :func:`matplotlib.pyplot.imshow` as extent.
            For the slider axes a [min, max] pair can be specified or an array the
            same length as the axis which will provide all values for that slider.
            If None is specified for an axis then the array indices will be used
            for that axis.

        """
        # For convenience in inserting dummy variables later, ensure
        # plot_axis_indices are all positive.
        plot_axis_indices = [i if i >= 0 else self.data.ndim + i for i in plot_axis_indices]
        # If axes kwargs not set by user, set them as list of Nones for
        # each axis for consistent behaviour.
        if axes_coordinates is None:
            axes_coordinates = [None] * self.data.ndim
        if axes_units is None:
            axes_units = [None] * self.data.ndim
        # If data_unit set, convert data to that unit
        if data_unit is None:
            data = self.data
        else:
            data = (self.data * self.unit).to(data_unit).value
        # Combine data values with mask.
        data = np.ma.masked_array(data, self.mask)
        # If axes_coordinates not provided generate an ImageAnimatorWCS plot
        # using NDCube's wcs object.
        if (axes_coordinates[plot_axis_indices[0]] is None and
                axes_coordinates[plot_axis_indices[1]] is None):
            # If there are missing axes in WCS object, add corresponding dummy axes to data.
            if data.ndim < self.wcs.naxis:
                new_shape = list(data.shape)
                for i in np.arange(self.wcs.naxis)[self.missing_axis[::-1]]:
                    new_shape.insert(i, 1)
                    # Also insert dummy coordinates and units.
                    axes_coordinates.insert(i, None)
                    axes_units.insert(i, None)
                    # Iterate plot_axis_indices if neccessary
                    for j, pai in enumerate(plot_axis_indices):
                        if pai >= i:
                            plot_axis_indices[j] = plot_axis_indices[j] + 1
                # Reshape data
                data = data.reshape(new_shape)
            # Generate plot
            ax = ImageAnimatorWCS(data, wcs=self.wcs, image_axes=plot_axis_indices,
                                  unit_x_axis=axes_units[plot_axis_indices[0]],
                                  unit_y_axis=axes_units[plot_axis_indices[1]],
                                  axis_ranges=axes_coordinates, **kwargs)
        # If one of the plot axes is set manually, produce a basic ImageAnimator object.
        else:
            new_axes_coordinates, new_axes_units, default_labels = \
              self._derive_axes_coordinates(axes_coordinates, axes_units)
            # If axis labels not set by user add to kwargs.
            ax = ImageAnimator(data, image_axes=plot_axis_indices,
                               axis_ranges=new_axes_coordinates, **kwargs)
        return ax

    def _animate_cube_1D(self, plot_axis_index=-1, axes_coordinates=None,
                         axes_units=None, data_unit=None, **kwargs):
        """Animates an axis of a cube as a line plot with sliders for other axes."""
        if axes_coordinates is None:
            axes_coordinates = [None] * self.data.ndim
        if axes_units is None:
            axes_units = [None] * self.data.ndim
        # Get real world axis values along axis to be plotted and enter into axes_ranges kwarg.
        if axes_coordinates[plot_axis_index] is None:
            xname = self.world_axis_physical_types[plot_axis_index]
            xdata = self.axis_world_coords(plot_axis_index)
        elif isinstance(axes_coordinates[plot_axis_index], str):
            xname = axes_coordinates[plot_axis_index]
            xdata = self.extra_coords[xname]["value"]
        else:
            xname = ""
            xdata = axes_coordinates[plot_axis_index]
        # Change x data to desired units it set by user.
        if isinstance(xdata, u.Quantity):
            if axes_units[plot_axis_index] is None:
                unit_x_axis = xdata.unit
            else:
                unit_x_axis = axes_units[plot_axis_index]
                xdata = xdata.to(unit_x_axis).value
        else:
            if axes_units[plot_axis_index] is not None:
                raise TypeError(INVALID_UNIT_SET_MESSAGE)
            else:
                unit_x_axis = None
        # Put xdata back into axes_coordinates as a masked array.
        axes_coordinates[plot_axis_index] = xdata
        # Set default x label
        default_xlabel = "{0} [{1}]".format(xname, unit_x_axis)
        # Derive y axis data
        if data_unit is None:
            data = self.data
            data_unit = self.unit
        else:
            if self.unit is None:
                raise TypeError("NDCube.unit is None.  Must be an astropy.units.unit or "
                                "valid unit string in order to set data_unit.")
            else:
                data = (self.data * self.unit).to(data_unit).value
        # Combine data with mask
        #data = np.ma.masked_array(data, self.mask)
        # Set default y label
        default_ylabel = "Data [{0}]".format(unit_x_axis)
        # Initiate line animator object.
        ax = LineAnimator(data, plot_axis_index=plot_axis_index, axis_ranges=axes_coordinates,
                          xlabel=default_xlabel,
                          ylabel="Data [{0}]".format(data_unit), **kwargs)
        return ax

    def _derive_axes_coordinates(self, axes_coordinates, axes_units):
        new_axes_coordinates = []
        new_axes_units = []
        default_labels = []
        default_label_text = ""
        for i, axis_coordinate in enumerate(axes_coordinates):
            # If axis coordinate is None, derive axis values from WCS.
            if axis_coordinate is None:
                # N.B. This assumes axes are independent.  Fix this before merging!!!
                new_axis_coordinate = self.axis_world_coords(i)
                axis_label_text = self.world_axis_physical_types[i]
            elif isinstance(axis_coordinate, str):
                # If axis coordinate is a string, derive axis values from
                # corresponding extra coord.
                new_axis_coordinate = self.extra_coords[axis_coordinate]["value"]
                axis_label_text = axis_coordinate
            else:
                # Else user must have manually set the axis coordinates.
                new_axis_coordinate = axis_coordinate
                axis_label_text = default_label_text
            # If axis coordinate is a Quantity, convert to unit supplied by user.
            if isinstance(new_axis_coordinate, u.Quantity):
                if axes_units[i] is None:
                    new_axis_unit = new_axis_coordinate.unit
                    new_axis_coordinate = new_axis_coordinate.value
                else:
                    new_axis_unit = axes_units[i]
                    new_axis_coordinate = new_axis_coordinate.to(new_axis_unit).value
            elif isinstance(new_axis_coordinate[0], datetime.datetime):
                axis_label_text = "{0}/sec since {1}".format(
                    axis_label_text, new_axis_coordinate[0])
                new_axis_coordinate = np.array([(t-new_axis_coordinate[0]).total_seconds()
                                                for t in new_axis_coordinate])
                new_axis_unit = u.s
            else:
                if axes_units[i] is None:
                    new_axis_unit = None
                else:
                    raise TypeError(INVALID_UNIT_SET_MESSAGE)
            # Derive default axis label
            if type(new_axis_coordinate[0]) is datetime.datetime:
                if axis_label_text == default_label_text:
                    default_label = "{0}".format(new_axis_coordinate[0].strftime("%Y/%m/%d %H:%M"))
                else:
                    default_label = "{0} [{1}]".format(
                        axis_label_text, new_axis_coordinate[0].strftime("%Y/%m/%d %H:%M"))
            else:
                default_label = "{0} [{1}]".format(axis_label_text, new_axis_unit)
            # Append new coordinates, units and labels to output list.
            new_axes_coordinates.append(new_axis_coordinate)
            new_axes_units.append(new_axis_unit)
            default_labels.append(default_label)
        return new_axes_coordinates, new_axes_units, default_labels


def _support_101_plot_API(plot_axis_indices, axes_coordinates, axes_units, data_unit, kwargs):
    """Check if user has used old API and convert it to new API."""
    # Get old API variable values.
    image_axes = kwargs.pop("image_axes", None)
    axis_ranges = kwargs.pop("axis_ranges", None)
    unit_x_axis = kwargs.pop("unit_x_axis", None)
    unit_y_axis = kwargs.pop("unit_y_axis", None)
    unit = kwargs.pop("unit", None)
    # Check if conflicting new and old API values have been set.
    # If not, set new API using old API and raise deprecation warning.
    if image_axes is not None:
        variable_names = ("image_axes", "plot_axis_indices")
        _raise_101_API_deprecation_warning(*variable_names)
        if plot_axis_indices is None:
            plot_axis_indices = image_axes
        else:
            _raise_API_error(*variable_names)
    if axis_ranges is not None:
        variable_names = ("axis_ranges", "axes_coordinates")
        _raise_101_API_deprecation_warning(*variable_names)
        if axes_coordinates is None:
            axes_coordinates = axis_ranges
        else:
            _raise_API_error(*variable_names)
    if (unit_x_axis is not None or unit_y_axis is not None) and axes_units is not None:
        _raise_API_error("unit_x_axis and/or unit_y_axis", "axes_units")
    if axes_units is None:
        variable_names = ("unit_x_axis and unit_y_axis", "axes_units")
        if unit_x_axis is not None:
            _raise_101_API_deprecation_warning(*variable_names)
            if len(plot_axis_indices) == 1:
                axes_units = unit_x_axis
            elif len(plot_axis_indices) == 2:
                if unit_y_axis is None:
                    axes_units = [unit_x_axis, None]
                else:
                    axes_units = [unit_x_axis, unit_y_axis]
            else:
                raise ValueError("Length of image_axes must be less than 3.")
        else:
            if unit_y_axis is not None:
                _raise_101_API_deprecation_warning(*variable_names)
                axes_units = [None, unit_y_axis]
    if unit is not None:
        variable_names = ("unit", "data_unit")
        _raise_101_API_deprecation_warning(*variable_names)
        if data_unit is None:
            data_unit = unit
        else:
            _raise_API_error(*variable_names)
    # Return values of new API
    return plot_axis_indices, axes_coordinates, axes_units, data_unit, kwargs


def _raise_API_error(old_name, new_name):
    raise ValueError(
        "Conflicting inputs: {0} (old API) cannot be set if {1} (new API) is set".format(
            old_name, new_name))

def _raise_101_API_deprecation_warning(old_name, new_name):
    warn("{0} is deprecated and will not be supported in version 2.0.  It will be replaced by {1}.  See docstring.".format(old_name, new_name), DeprecationWarning)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Utilities for generating the version string for Astropy (or an affiliated
package) and the version.py module, which contains version info for the
package.

Within the generated astropy.version module, the `major`, `minor`, and `bugfix`
variables hold the respective parts of the version number (bugfix is '0' if
absent). The `release` variable is True if this is a release, and False if this
is a development version of astropy. For the actual version string, use::

    from astropy.version import version

or::

    from astropy import __version__

"""

from __future__ import division

import datetime
import imp
import os
import pkgutil
import sys
import time

from distutils import log
from importlib import invalidate_caches

import pkg_resources

from . import git_helpers
from .distutils_helpers import is_distutils_display_option


def _version_split(version):
    """
    Split a version string into major, minor, and bugfix numbers.  If any of
    those numbers are missing the default is zero.  Any pre/post release
    modifiers are ignored.

    Examples
    ========
    >>> _version_split('1.2.3')
    (1, 2, 3)
    >>> _version_split('1.2')
    (1, 2, 0)
    >>> _version_split('1.2rc1')
    (1, 2, 0)
    >>> _version_split('1')
    (1, 0, 0)
    >>> _version_split('')
    (0, 0, 0)
    """

    parsed_version = pkg_resources.parse_version(version)

    if hasattr(parsed_version, 'base_version'):
        # New version parsing for setuptools >= 8.0
        if parsed_version.base_version:
            parts = [int(part)
                     for part in parsed_version.base_version.split('.')]
        else:
            parts = []
    else:
        parts = []
        for part in parsed_version:
            if part.startswith('*'):
                # Ignore any .dev, a, b, rc, etc.
                break
            parts.append(int(part))

    if len(parts) < 3:
        parts += [0] * (3 - len(parts))

    # In principle a version could have more parts (like 1.2.3.4) but we only
    # support <major>.<minor>.<micro>
    return tuple(parts[:3])


# This is used by setup.py to create a new version.py - see that file for
# details. Note that the imports have to be absolute, since this is also used
# by affiliated packages.
_FROZEN_VERSION_PY_TEMPLATE = """
# Autogenerated by {packagetitle}'s setup.py on {timestamp!s} UTC
from __future__ import unicode_literals
import datetime

{header}

major = {major}
minor = {minor}
bugfix = {bugfix}

version_info = (major, minor, bugfix)

release = {rel}
timestamp = {timestamp!r}
debug = {debug}

astropy_helpers_version = "{ahver}"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
"""[1:]


_FROZEN_VERSION_PY_WITH_GIT_HEADER = """
{git_helpers}


_packagename = "{packagename}"
_last_generated_version = "{verstr}"
_last_githash = "{githash}"

# Determine where the source code for this module
# lives.  If __file__ is not a filesystem path then
# it is assumed not to live in a git repo at all.
if _get_repo_path(__file__, levels=len(_packagename.split('.'))):
    version = update_git_devstr(_last_generated_version, path=__file__)
    githash = get_git_devstr(sha=True, show_warning=False,
                             path=__file__) or _last_githash
else:
    # The file does not appear to live in a git repo so don't bother
    # invoking git
    version = _last_generated_version
    githash = _last_githash
"""[1:]


_FROZEN_VERSION_PY_STATIC_HEADER = """
version = "{verstr}"
githash = "{githash}"
"""[1:]


def _get_version_py_str(packagename, version, githash, release, debug,
                        uses_git=True):
    try:
        from astropy_helpers import __version__ as ahver
    except ImportError:
        ahver = "unknown"

    epoch = int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))
    timestamp = datetime.datetime.utcfromtimestamp(epoch)
    major, minor, bugfix = _version_split(version)

    if packagename.lower() == 'astropy':
        packagetitle = 'Astropy'
    else:
        packagetitle = 'Astropy-affiliated package ' + packagename

    header = ''

    if uses_git:
        header = _generate_git_header(packagename, version, githash)
    elif not githash:
        # _generate_git_header will already generate a new git has for us, but
        # for creating a new version.py for a release (even if uses_git=False)
        # we still need to get the githash to include in the version.py
        # See https://github.com/astropy/astropy-helpers/issues/141
        githash = git_helpers.get_git_devstr(sha=True, show_warning=True)

    if not header:  # If _generate_git_header fails it returns an empty string
        header = _FROZEN_VERSION_PY_STATIC_HEADER.format(verstr=version,
                                                         githash=githash)

    return _FROZEN_VERSION_PY_TEMPLATE.format(packagetitle=packagetitle,
                                              timestamp=timestamp,
                                              header=header,
                                              major=major,
                                              minor=minor,
                                              bugfix=bugfix,
                                              ahver=ahver,
                                              rel=release, debug=debug)


def _generate_git_header(packagename, version, githash):
    """
    Generates a header to the version.py module that includes utilities for
    probing the git repository for updates (to the current git hash, etc.)
    These utilities should only be available in development versions, and not
    in release builds.

    If this fails for any reason an empty string is returned.
    """

    loader = pkgutil.get_loader(git_helpers)
    source = loader.get_source(git_helpers.__name__) or ''
    source_lines = source.splitlines()
    if not source_lines:
        log.warn('Cannot get source code for astropy_helpers.git_helpers; '
                 'git support disabled.')
        return ''

    idx = 0
    for idx, line in enumerate(source_lines):
        if line.startswith('# BEGIN'):
            break
    git_helpers_py = '\n'.join(source_lines[idx + 1:])

    verstr = version

    new_githash = git_helpers.get_git_devstr(sha=True, show_warning=False)

    if new_githash:
        githash = new_githash

    return _FROZEN_VERSION_PY_WITH_GIT_HEADER.format(
                git_helpers=git_helpers_py, packagename=packagename,
                verstr=verstr, githash=githash)


def generate_version_py(packagename, version, release=None, debug=None,
                        uses_git=True, srcdir='.'):
    """Regenerate the version.py module if necessary."""

    try:
        version_module = get_pkg_version_module(packagename)

        try:
            last_generated_version = version_module._last_generated_version
        except AttributeError:
            last_generated_version = version_module.version

        try:
            last_githash = version_module._last_githash
        except AttributeError:
            last_githash = version_module.githash

        current_release = version_module.release
        current_debug = version_module.debug
    except ImportError:
        version_module = None
        last_generated_version = None
        last_githash = None
        current_release = None
        current_debug = None

    if release is None:
        # Keep whatever the current value is, if it exists
        release = bool(current_release)

    if debug is None:
        # Likewise, keep whatever the current value is, if it exists
        debug = bool(current_debug)

    package_srcdir = os.path.join(srcdir, *packagename.split('.'))
    version_py = os.path.join(package_srcdir, 'version.py')

    if (last_generated_version != version or current_release != release or
            current_debug != debug):
        if '-q' not in sys.argv and '--quiet' not in sys.argv:
            log.set_threshold(log.INFO)

        if is_distutils_display_option():
            # Always silence unnecessary log messages when display options are
            # being used
            log.set_threshold(log.WARN)

        log.info('Freezing version number to {0}'.format(version_py))

        with open(version_py, 'w') as f:
            # This overwrites the actual version.py
            f.write(_get_version_py_str(packagename, version, last_githash,
                                        release, debug, uses_git=uses_git))

        invalidate_caches()

        if version_module:
            imp.reload(version_module)


def get_pkg_version_module(packagename, fromlist=None):
    """Returns the package's .version module generated by
    `astropy_helpers.version_helpers.generate_version_py`.  Raises an
    ImportError if the version module is not found.

    If ``fromlist`` is an iterable, return a tuple of the members of the
    version module corresponding to the member names given in ``fromlist``.
    Raises an `AttributeError` if any of these module members are not found.
    """

    if not fromlist:
        # Due to a historical quirk of Python's import implementation,
        # __import__ will not return submodules of a package if 'fromlist' is
        # empty.
        # TODO: For Python 3.1 and up it may be preferable to use importlib
        # instead of the __import__ builtin
        return __import__(packagename + '.version', fromlist=[''])
    else:
        mod = __import__(packagename + '.version', fromlist=fromlist)
        return tuple(getattr(mod, member) for member in fromlist)
